@@ -23,7 +23,6 @@ import (
 	"net/http"
 	"os"
 	"path"
-	"regexp"
 	"strconv"
 	"strings"
 	"sync"
@@ -34,7 +33,6 @@ import (
 	"github.com/appc/docker2aci/lib/internal/types"
 	"github.com/appc/docker2aci/lib/internal/typesV2"
 	"github.com/appc/docker2aci/lib/internal/util"
-	"github.com/appc/docker2aci/pkg/log"
 	"github.com/appc/spec/schema"
 	"github.com/coreos/pkg/progressutil"
 )
@@ -43,8 +41,6 @@ const (
 	defaultIndexURL = "registry-1.docker.io"
 )
 
-var validHex = regexp.MustCompile(`^([a-f0-9]{64})$`)
-
 // A manifest conforming to the docker v2.1 spec
 type v2Manifest struct {
 	Name     string `json:"name"`
@@ -58,7 +54,7 @@ type v2Manifest struct {
 	Signature []byte `json:"signature"`
 }
 
-func (rb *RepositoryBackend) getImageInfoV2(dockerURL *types.ParsedDockerURL) ([]string, *types.ParsedDockerURL, error) {
+func (rb *RepositoryBackend) getImageInfoV2(dockerURL *common.ParsedDockerURL) ([]string, *common.ParsedDockerURL, error) {
 	layers, err := rb.getManifestV2(dockerURL)
 	if err != nil {
 		return nil, nil, err
@@ -67,7 +63,7 @@ func (rb *RepositoryBackend) getImageInfoV2(dockerURL *types.ParsedDockerURL) ([
 	return layers, dockerURL, nil
 }
 
-func (rb *RepositoryBackend) buildACIV2(layerIDs []string, dockerURL *types.ParsedDockerURL, outputDir string, tmpBaseDir string, compression common.Compression) ([]string, []*schema.ImageManifest, error) {
+func (rb *RepositoryBackend) buildACIV2(layerIDs []string, dockerURL *common.ParsedDockerURL, outputDir string, tmpBaseDir string, compression common.Compression) ([]string, []*schema.ImageManifest, error) {
 	_, isVersion22 := rb.imageV2Manifests[*dockerURL]
 	if isVersion22 {
 		return rb.buildACIV22(layerIDs, dockerURL, outputDir, tmpBaseDir, compression)
@@ -75,7 +71,7 @@ func (rb *RepositoryBackend) buildACIV2(layerIDs []string, dockerURL *types.Pars
 	return rb.buildACIV21(layerIDs, dockerURL, outputDir, tmpBaseDir, compression)
 }
 
-func (rb *RepositoryBackend) buildACIV21(layerIDs []string, dockerURL *types.ParsedDockerURL, outputDir string, tmpBaseDir string, compression common.Compression) ([]string, []*schema.ImageManifest, error) {
+func (rb *RepositoryBackend) buildACIV21(layerIDs []string, dockerURL *common.ParsedDockerURL, outputDir string, tmpBaseDir string, compression common.Compression) ([]string, []*schema.ImageManifest, error) {
 	layerFiles := make([]*os.File, len(layerIDs))
 	layerDatas := make([]types.DockerImageData, len(layerIDs))
 
@@ -91,6 +87,9 @@ func (rb *RepositoryBackend) buildACIV21(layerIDs []string, dockerURL *types.Par
 	closers := make([]io.ReadCloser, len(layerIDs))
 	var wg sync.WaitGroup
 	for i, layerID := range layerIDs {
+		if err := common.ValidateLayerId(layerID); err != nil {
+			return nil, nil, err
+		}
 		wg.Add(1)
 		errChan := make(chan error, 1)
 		errChannels = append(errChannels, errChan)
@@ -160,8 +159,8 @@ func (rb *RepositoryBackend) buildACIV21(layerIDs []string, dockerURL *types.Par
 	var aciManifests []*schema.ImageManifest
 	var curPwl []string
 	for i := len(layerIDs) - 1; i >= 0; i-- {
-		log.Debug("Generating layer ACI...")
-		aciPath, aciManifest, err := internal.GenerateACI(i, layerDatas[i], dockerURL, outputDir, layerFiles[i], curPwl, compression)
+		rb.debug.Println("Generating layer ACI...")
+		aciPath, aciManifest, err := internal.GenerateACI(i, layerDatas[i], dockerURL, outputDir, layerFiles[i], curPwl, compression, rb.debug)
 		if err != nil {
 			return nil, nil, fmt.Errorf("error generating ACI: %v", err)
 		}
@@ -182,7 +181,7 @@ type layer struct {
 	err    error
 }
 
-func (rb *RepositoryBackend) buildACIV22(layerIDs []string, dockerURL *types.ParsedDockerURL, outputDir string, tmpBaseDir string, compression common.Compression) ([]string, []*schema.ImageManifest, error) {
+func (rb *RepositoryBackend) buildACIV22(layerIDs []string, dockerURL *common.ParsedDockerURL, outputDir string, tmpBaseDir string, compression common.Compression) ([]string, []*schema.ImageManifest, error) {
 	layerFiles := make([]*os.File, len(layerIDs))
 
 	tmpParentDir, err := ioutil.TempDir(tmpBaseDir, "docker2aci-")
@@ -195,6 +194,9 @@ func (rb *RepositoryBackend) buildACIV22(layerIDs []string, dockerURL *types.Par
 
 	resultChan := make(chan layer, len(layerIDs))
 	for i, layerID := range layerIDs {
+		if err := common.ValidateLayerId(layerID); err != nil {
+			return nil, nil, err
+		}
 		// https://github.com/golang/go/wiki/CommonMistakes
 		i := i // golang--
 		layerID := layerID
@@ -256,7 +258,7 @@ func (rb *RepositoryBackend) buildACIV22(layerIDs []string, dockerURL *types.Par
 	var curPwl []string
 	var i int
 	for i = 0; i < len(layerIDs)-1; i++ {
-		log.Debug("Generating layer ACI...")
+		rb.debug.Println("Generating layer ACI...")
 		aciPath, aciManifest, err := internal.GenerateACI22LowerLayer(dockerURL, layerIDs[i], outputDir, layerFiles[i], curPwl, compression)
 		if err != nil {
 			return nil, nil, fmt.Errorf("error generating ACI: %v", err)
@@ -265,8 +267,8 @@ func (rb *RepositoryBackend) buildACIV22(layerIDs []string, dockerURL *types.Par
 		aciManifests = append(aciManifests, aciManifest)
 		curPwl = aciManifest.PathWhitelist
 	}
-	log.Debug("Generating layer ACI...")
-	aciPath, aciManifest, err := internal.GenerateACI22TopLayer(dockerURL, rb.imageConfigs[*dockerURL], layerIDs[i], outputDir, layerFiles[i], curPwl, compression, aciManifests)
+	rb.debug.Println("Generating layer ACI...")
+	aciPath, aciManifest, err := internal.GenerateACI22TopLayer(dockerURL, rb.imageConfigs[*dockerURL], layerIDs[i], outputDir, layerFiles[i], curPwl, compression, aciManifests, rb.debug)
 	if err != nil {
 		return nil, nil, fmt.Errorf("error generating ACI: %v", err)
 	}
@@ -276,7 +278,7 @@ func (rb *RepositoryBackend) buildACIV22(layerIDs []string, dockerURL *types.Par
 	return aciLayerPaths, aciManifests, nil
 }
 
-func (rb *RepositoryBackend) getManifestV2(dockerURL *types.ParsedDockerURL) ([]string, error) {
+func (rb *RepositoryBackend) getManifestV2(dockerURL *common.ParsedDockerURL) ([]string, error) {
 	var reference string
 	if dockerURL.Digest != "" {
 		reference = dockerURL.Digest
@@ -305,7 +307,7 @@ func (rb *RepositoryBackend) getManifestV2(dockerURL *types.ParsedDockerURL) ([]
 	defer res.Body.Close()
 
 	if res.StatusCode != http.StatusOK {
-		return nil, fmt.Errorf("unexpected http code: %d, URL: %s", res.StatusCode, req.URL)
+		return nil, &httpStatusErr{res.StatusCode, req.URL}
 	}
 
 	switch res.Header.Get("content-type") {
@@ -317,7 +319,7 @@ func (rb *RepositoryBackend) getManifestV2(dockerURL *types.ParsedDockerURL) ([]
 	return rb.getManifestV21(dockerURL, res)
 }
 
-func (rb *RepositoryBackend) getManifestV21(dockerURL *types.ParsedDockerURL, res *http.Response) ([]string, error) {
+func (rb *RepositoryBackend) getManifestV21(dockerURL *common.ParsedDockerURL, res *http.Response) ([]string, error) {
 	manblob, err := ioutil.ReadAll(res.Body)
 	if err != nil {
 		return nil, err
@@ -334,7 +336,7 @@ func (rb *RepositoryBackend) getManifestV21(dockerURL *types.ParsedDockerURL, re
 		return nil, fmt.Errorf("name doesn't match what was requested, expected: %s, downloaded: %s", dockerURL.ImageName, manifest.Name)
 	}
 
-	if manifest.Tag != dockerURL.Tag {
+	if dockerURL.Tag != "" && manifest.Tag != dockerURL.Tag {
 		return nil, fmt.Errorf("tag doesn't match what was requested, expected: %s, downloaded: %s", dockerURL.Tag, manifest.Tag)
 	}
 
@@ -358,7 +360,7 @@ func (rb *RepositoryBackend) getManifestV21(dockerURL *types.ParsedDockerURL, re
 	return layers, nil
 }
 
-func (rb *RepositoryBackend) getManifestV22(dockerURL *types.ParsedDockerURL, res *http.Response) ([]string, error) {
+func (rb *RepositoryBackend) getManifestV22(dockerURL *common.ParsedDockerURL, res *http.Response) ([]string, error) {
 	manblob, err := ioutil.ReadAll(res.Body)
 	if err != nil {
 		return nil, err
@@ -389,7 +391,7 @@ func (rb *RepositoryBackend) getManifestV22(dockerURL *types.ParsedDockerURL, re
 	return layers, nil
 }
 
-func (rb *RepositoryBackend) getConfigV22(dockerURL *types.ParsedDockerURL, configDigest string) error {
+func (rb *RepositoryBackend) getConfigV22(dockerURL *common.ParsedDockerURL, configDigest string) error {
 	url := rb.schema + path.Join(dockerURL.IndexURL, "v2", dockerURL.ImageName, "blobs", configDigest)
 	req, err := http.NewRequest("GET", url, nil)
 	if err != nil {
@@ -436,7 +438,7 @@ func fixManifestLayers(manifest *v2Manifest) error {
 		}
 
 		imgs[i] = img
-		if err := validateV1ID(img.ID); err != nil {
+		if err := common.ValidateLayerId(img.ID); err != nil {
 			return err
 		}
 	}
@@ -471,14 +473,7 @@ func fixManifestLayers(manifest *v2Manifest) error {
 	return nil
 }
 
-func validateV1ID(id string) error {
-	if ok := validHex.MatchString(id); !ok {
-		return fmt.Errorf("image ID %q is invalid", id)
-	}
-	return nil
-}
-
-func (rb *RepositoryBackend) getLayerV2(layerID string, dockerURL *types.ParsedDockerURL, tmpDir string, copier *progressutil.CopyProgressPrinter) (*os.File, io.ReadCloser, error) {
+func (rb *RepositoryBackend) getLayerV2(layerID string, dockerURL *common.ParsedDockerURL, tmpDir string, copier *progressutil.CopyProgressPrinter) (*os.File, io.ReadCloser, error) {
 	var (
 		err error
 		res *http.Response
@@ -493,7 +488,7 @@ func (rb *RepositoryBackend) getLayerV2(layerID string, dockerURL *types.ParsedD
 
 	accepting := []string{
 		typesV2.MediaTypeDockerV22RootFS,
-		typesV2.MediaTypeOCIRootFS,
+		typesV2.MediaTypeOCILayer,
 	}
 
 	res, err = rb.makeRequest(req, dockerURL.ImageName, accepting)
@@ -524,7 +519,7 @@ func (rb *RepositoryBackend) getLayerV2(layerID string, dockerURL *types.ParsedD
 	}
 
 	if res.StatusCode != http.StatusOK {
-		return nil, nil, fmt.Errorf("HTTP code: %d. URL: %s", res.StatusCode, req.URL)
+		return nil, nil, &httpStatusErr{res.StatusCode, req.URL}
 	}
 
 	var in io.Reader
@@ -642,7 +637,7 @@ func (rb *RepositoryBackend) makeRequest(req *http.Request, repo string, acceptH
 	case http.StatusOK:
 		break
 	default:
-		return nil, fmt.Errorf("unexpected http code: %d, URL: %s", res.StatusCode, authReq.URL)
+		return nil, &httpStatusErr{res.StatusCode, authReq.URL}
 	}
 
 	tokenBlob, err := ioutil.ReadAll(res.Body)
