@@ -17,6 +17,8 @@ fi
 export GO15VENDOREXPERIMENT=1
 export GOPATH=${DIR}/gopath
 
+go vet ./pkg/...
+go vet ./lib/...
 go test -v ${REPO_PATH}/lib/tests
 
 DOCKER2ACI=../bin/docker2aci
@@ -45,6 +47,12 @@ RKT=$(which rkt)
 
 DOCKER_STORAGE_BACKEND=$(sudo docker info|grep '^Storage Driver:'|sed 's/Storage Driver: //')
 
+for i in $(find . -maxdepth 1 -type d -name 'fixture-test*') ; do
+	  TESTNAME=$(basename $i)
+	  echo "### Test case ${TESTNAME}..."
+	  $TESTDIR/${TESTNAME}/check.sh "${TESTDIR}" "${TESTNAME}"
+done
+
 for i in $(find . -maxdepth 1 -type d -name 'test-*') ; do
 	TESTNAME=$(basename $i)
 	echo "### Test case ${TESTNAME}: build..."
@@ -81,4 +89,3 @@ for i in $(find . -maxdepth 1 -type d -name 'test-*') ; do
 
 	sudo docker rmi $PREFIX/${TESTNAME}
 done
-
